/*
 *  2.1.    a*b+c*d.    
 * ,    .    .
 */

#include <stdio.h>
#include <stdlib.h>

#define LINE_BUFFER_LENGTH  256
#define STACK_MEM_CHUNK     16

const char *strGreeting="Reverse polish integer calculator.\n\
Enter expression(s) to calculate (Ctrl-D to quit):\n";

const char *strStackCleared="stack cleared\n";
const char *strErrMem="error: not enough memory\n";
const char *strErrStackEmpty="error: stack empty\n";
const char *strErrUnknownOperation="error: unknown operation `%c'\n";
const char *strErrUnknownCommand="error: unknown command `%c'\n";

char *line;

void ParseLine(char *line);
int  ParseCommand(char **ptr);
int  PopOneOp(long *ptrOp1);
int  PopTwoOps(long *ptrOp1, long *ptrOp2);
int  PushAndPrintResult(long num);

int  Push(long Value);
int  Pop(long *Value);
int  GetTop(long *Value);
void ClearStack(void);
void PrintStack(void);
void FreeStack(void);



int main()
{
  fputs(strGreeting, stdout);

  if ( !(line = malloc(LINE_BUFFER_LENGTH)) ) {
    fputs(strErrMem, stderr);
    return -1;
  }

  while ( fgets(line, LINE_BUFFER_LENGTH, stdin) ) {
    ParseLine(line);
  }

  FreeStack();

  return 0;
}



void ParseLine(char *line)
{
  char *ptr;
  long op1, op2;
  int  sgn, BailOut;

  ptr = line;

  BailOut = 0;
  while ( *ptr && !BailOut ) {
    if ( isspace(*ptr) ) {
      /* Skip spaces */
      ptr++;

    } else if ( isalpha(*ptr) ) {
      /* Parse out the command */
      if ( !ParseCommand(&ptr) )
        BailOut = -1;

    } else if ( isdigit(*ptr) || *ptr == '_' ) {
      /* Parse out the number */
      if ( *ptr == '_' ) {
        sgn = -1;
        ptr++;
      } else
        sgn = 0;
      op1 = strtol(ptr, &ptr, 0);
      if ( sgn )
        op1 = -op1;
      if ( !Push(op1) ) {
        fputs(strErrMem, stderr);
        BailOut = -1;
      }

    } else {
      /* Parse out the operation */
      switch ( *ptr ) {
      case '+':
        (BailOut = PopTwoOps(&op1, &op2)) ||
          (BailOut = PushAndPrintResult(op1+op2));
        break;
      case '-':
        (BailOut = PopTwoOps(&op1, &op2)) ||
          (BailOut = PushAndPrintResult(op1-op2));
        break;
      case '*':
        (BailOut = PopTwoOps(&op1, &op2)) ||
          (BailOut = PushAndPrintResult(op1*op2));
        break;
      case '/':
        (BailOut = PopTwoOps(&op1, &op2)) ||
          (BailOut = PushAndPrintResult(op1/op2));
        break;
      case '\%': /* modulo */
        (BailOut = PopTwoOps(&op1, &op2)) ||
          (BailOut = PushAndPrintResult(op1%op2));
        break;
      case '$': /* absolute value */
        (BailOut = PopOneOp(&op1)) ||
          ( BailOut = PushAndPrintResult( (op1 < 0) ? -op1 : op1 ) );
        break;
      case '~': /* negate */
        (BailOut = PopOneOp(&op1)) || ( BailOut = PushAndPrintResult(-op1) );
        break;
      default:
        fprintf(stderr, strErrUnknownOperation, *ptr);
        BailOut = -1;
        break;
      }
      ptr++;
    }
  }

  return;
}

/* Returns -1 on success, 0 if command is unknown */ 
int ParseCommand(char **ptr)
{
  long num, op1;
  int ret;

  ret = -1;
  switch (**ptr) {
  case 't': /* Print top stack value */
    if ( !GetTop(&num) )
      fprintf(stderr, strErrStackEmpty);
    else
      printf("%ld\n", num);
    break;
  case 's': /* Print entire stack */
    PrintStack();
    break;
  case 'p': /* Pop off top stack value */
    if ( !Pop(&num) )
      fprintf(stderr, strErrStackEmpty);
    else
      printf("%ld\n", num);
    break;
  case 'c': /* Clear entire stack */
    ClearStack();
    fputs(strStackCleared, stdout);
    break;
  case 'e': /* Exchange two top values in the stack */
    PopTwoOps(&num, &op1) || PushAndPrintResult(op1) || PushAndPrintResult(num);
    break;
  case 'd': /* Duplicate top value */
    if ( !GetTop(&num) )
      fprintf(stderr, strErrStackEmpty);
    else
      PushAndPrintResult(num);
    break;
  default:
    fprintf(stderr, strErrUnknownCommand, **ptr);
    ret = 0;
    break;
  }

  (*ptr)++;

  return ret;
}

int PopOneOp(long *ptrOp1)
{
  if ( !Pop(ptrOp1) ) {
    fputs(strErrStackEmpty, stderr);
    return -1;
  }

  return 0;
}

int PopTwoOps(long *ptrOp1, long *ptrOp2)
{
  if ( !Pop(ptrOp2) || !Pop(ptrOp1) ) {
    fputs(strErrStackEmpty, stderr);
    return -1;
  }

  return 0;
}

int PushAndPrintResult(long num)
{
  if ( !Push(num) ) {
    fputs(strErrMem, stderr);
    return -1;
  }

  printf("%ld\n", num);
  return 0;
}



static long *Stack = NULL;
static int StackTop = -1;
static int StackLen = 0;

/* Returns length of stack on success, 0 on failure to grow stack */
int Push(long Value)
{
  long *ptr;

  /* Grow stack if necessary */
  if ( ++StackTop >= StackLen ) {
    StackLen += STACK_MEM_CHUNK;
    if ( !(ptr = (long *) realloc(Stack, StackLen * sizeof(*Stack))) ) {
      StackLen -= STACK_MEM_CHUNK;
      return 0;
    }
    Stack = ptr;
  }
  /* Push the value onto stack */
  *(Stack + StackTop) = Value;

  return StackTop + 1;
}

/* Returns -1 on success, 0 if stack is empty */
int Pop(long *Value)
{
  if ( StackTop >= 0 ) {
    *Value = *(Stack + StackTop--);
    return -1;
  }

  return 0;
}

/* Returns -1 on success, 0 if stack is empty */
int GetTop(long *Value)
{
  if ( StackTop >= 0 ) {
    *Value = *(Stack + StackTop);
    return -1;
  }

  return 0;
}

void ClearStack(void)
{
  StackTop = -1;
}

void PrintStack(void)
{
  int i;

  if (StackTop < 0) {
    fputs(strErrStackEmpty, stderr);
    return;
  }

  i = 0;
  while ( i <= StackTop )
    printf("%ld\n", *(Stack + i++));
}

void FreeStack(void)
{
  free(Stack);
  Stack = NULL;
  StackLen = 0;
  StackTop = -1;
}



/*
 * Local variables:
 * compile-command: "gcc 2_01.c"
 * End:
 */
