#include "fp.h"

#define STRING_MEM_CHUNK     16

static char *str = NULL;
static int   strLen = 0;

static const char ParenOpen[]  = "([{<";
static const char ParenClose[] = ")]}>";

#define PAREN_AVAILABLE (sizeof(ParenOpen) - 1)

static int   UnparseNode(tnodeptr ptrNode, int *pos, int PrevPriority, int *ParenDepth);
static int   StrOut(char *from, int *pos);
static int   CharOut(char c, int *pos);
static void  CharOut_Restart();
       char *GetUnparseBuffer(void);
       void  FreeUnparseBuffer(void);
static int   itoa(long value, int *pos);
static void  reverse(int start, int end);

char *UnparseTree(tnodeptr ptrRoot)
{
  int i = 0, depth = 0;

  /* Initialize variables */
  CharOut_Restart();

  if ( !ptrRoot )
    return NULL;

  /* Start unparsing of the tree */
  if ( !UnparseNode(ptrRoot, &i, -1, &depth) ) {
    FreeUnparseBuffer();
    return NULL;
  }

  /* Finish off the string with '\0' */
  if ( !CharOut('\0', &i) ) {
    FreeUnparseBuffer();
    return NULL;
  }

  return str;
}

static int UnparseNode(tnodeptr ptrNode, int *pos, int PrevPriority, int *ParenDepth)
{
  if ( !ptrNode )
    return 0;

  if ( ptrNode->type == NT_OPERATION ) {
    /****************************** OPERATION ******************************/
    int OpenParenPos;
    operationptr ptrOp = (operationptr) Operations+ptrNode->value;

    if ( ptrOp->ops == 1 ) {
      /********** This is unary operation **********/ 
      tnodeptr ptrNextNode;
      int InsertParens = 0, BranchParenDepth = 0;

      /* Sanity check */
      if ( !ptrNode->ptrLeft && !ptrNode->ptrRight ||
            ptrNode->ptrLeft &&  ptrNode->ptrRight    ) {
        fputs(strErrUnparsingTree, stderr);
        return 0;
      }
      /* Deal with left-or-right ambiguity */
      if ( ptrNode->ptrLeft )
        ptrNextNode = ptrNode->ptrLeft;
      else
        ptrNextNode = ptrNode->ptrRight;
      /* Output operation's prefix */
      if ( !StrOut(ptrOp->prefix, pos) )
        return 0;
      /* See if we need to insert parentheses */
      if ( ptrOp->priority & UNARY_PAREN_OP_ALWAYS ||
           ptrOp->priority & UNARY_PAREN_OP_NESTED &&
           ( ptrNextNode->type == NT_OPERATION ||
             ptrNextNode->type == NT_NUMBER && ptrNextNode->value < 0 ) )
        InsertParens = -1;
      if ( InsertParens )
        OpenParenPos = (*pos)++;
      /* Output subtree */
      if ( !UnparseNode(ptrNextNode, pos, -1, &BranchParenDepth) )
        return 0;
      *ParenDepth = BranchParenDepth;
      /* Output opening and closing parentheses, if needed */
      if ( InsertParens ) {
        (*ParenDepth)++;
        if ( BranchParenDepth >= PAREN_AVAILABLE )
          BranchParenDepth = PAREN_AVAILABLE - 1;
        if ( !CharOut(*(ParenOpen + BranchParenDepth), &OpenParenPos) ||
             !CharOut(*(ParenClose + BranchParenDepth), pos) )
          return 0;
      }
      /* Finally, output operation's postfix */
      if ( !StrOut(ptrOp->postfix, pos) )
        return 0;

    } else {
      /********** This is binary operation **********/ 
      int ThisPriority, i;
      int LeftBranchParenDepth = 0, RightBranchParenDepth = 0;
      int RightBranchOpenParenPos, ParenRightOp = 0;

      /* Sanity check */
      if ( !ptrNode->ptrLeft || !ptrNode->ptrRight ) {
        fputs(strErrUnparsingTree, stderr);
        return 0;
      }
      ThisPriority = (Operations+ptrNode->value)->priority;
      /* If this operation's priority is lower than that of the
         previous, we need to parenthesise its operands */
      if ( PrevPriority >= 0 && ThisPriority < PrevPriority )
        OpenParenPos = (*pos)++;
      /* Output left subtree */
      if ( !UnparseNode(ptrNode->ptrLeft,
                        pos,
                        ThisPriority,
                        &LeftBranchParenDepth) )
        return 0;
      /* Output the operation */
      if ( !StrOut( (Operations+ptrNode->value)->prefix, pos ) )
        return 0;
      /* If right operand is a negative number, or it is a unary
         operation of UNARY_PAREN_WHEN_RIGHT type of parenthesising,
         we need to parenthesise it */
      if ( ptrNode->ptrRight->type==NT_NUMBER && ptrNode->ptrRight->value<0 ||
           ptrNode->ptrRight->type == NT_OPERATION &&
             (Operations + ptrNode->ptrRight->value)->ops == 1 &&
             (Operations + ptrNode->ptrRight->value)->priority & UNARY_PAREN_WHEN_RIGHT ) {
        ParenRightOp = -1;
        RightBranchOpenParenPos = (*pos)++;
      }
      /* Output right subtree */
      if ( !UnparseNode(ptrNode->ptrRight, pos, ThisPriority, &RightBranchParenDepth) )
        return 0;
      /* Output parentheses around right operand, if needed */
      if ( ParenRightOp ) {
        i = RightBranchParenDepth++;
        if ( i >= PAREN_AVAILABLE )
          i = PAREN_AVAILABLE - 1;
        if ( !CharOut(*(ParenOpen + i), &RightBranchOpenParenPos) ||
             !CharOut(*(ParenClose + i), pos) )
          return 0;
      }
      /* Choose the larger of the parenthesis depths for this node's depth */
      if ( LeftBranchParenDepth > RightBranchParenDepth )
        i = LeftBranchParenDepth;
      else
        i = RightBranchParenDepth;
      *ParenDepth = i;
      /* Output parentheses around entire operation, if needed */
      if ( PrevPriority >= 0 && ThisPriority < PrevPriority ) {
        (*ParenDepth)++;
        if ( i >= PAREN_AVAILABLE )
          i = PAREN_AVAILABLE - 1;
        if ( !CharOut(*(ParenOpen + i), &OpenParenPos) ||
             !CharOut(*(ParenClose + i), pos) )
          return 0;
      }
    }

  } else {
    /****************************** NUMBER ******************************/
    if ( !itoa(ptrNode->value, pos) )
      return 0;
  }

  return -1;
}

/* Returns -1 on success, 0 on error (memory allocation failed) */
static int StrOut(char *from, int *pos)
{
  if ( !from )
    return -1;

  while ( *from ) {
    if ( *pos >= strLen ) {
      char *ptr;

      strLen += STRING_MEM_CHUNK;
      if ( !(ptr = (char *) realloc((void *) str, strLen)) ) {
        strLen -= STRING_MEM_CHUNK;
        fputs(strErrMem, stderr);

        return 0;
      }

      str = ptr;
    }

    *(str + (*pos)++) = *from++;
  }

  return -1;
}

/* Returns -1 on success, 0 on error (memory allocation failed) */
static int CharOut(char c, int *pos)
{
  if ( *pos >= strLen ) {
    char *ptr;

    strLen += STRING_MEM_CHUNK;
    if ( !(ptr = (char *) realloc((void *) str, strLen)) ) {
      strLen -= STRING_MEM_CHUNK;
      fputs(strErrMem, stderr);

      return 0;
    }

    str = ptr;
  }

  *(str + (*pos)++) = c;

  return -1;
}

static void CharOut_Restart(void)
{
  str = NULL;
  strLen = 0;
}

char *GetUnparseBuffer(void)
{
  return str;
}

void FreeUnparseBuffer(void)
{
  if ( str )
    free((void *) str);

  CharOut_Restart();
}

static int itoa(long value, int *pos)
{
  int start;

  if ( value < 0 ) {
    if ( !CharOut('-', pos) )
      return 0;
    value = -value;
  }

  start = *pos;
  do {
    if ( !CharOut(value % 10 + '0', pos) )
      return 0;
    value /= 10;
  } while ( value > 0 );

  reverse(start, *pos - 1);

  return -1;
}

static void reverse(int start, int end)
{
  char c;

  if ( !str )
    return;

  while ( start < end ) {
    c = *(str + start);
    *(str + start++) = *(str + end);
    *(str + end--) = c;
  }

  return;
}



/*
 * Local variables:
 * compile-command: "make -k all && echo && ./fp fp.in"
 * End:
 */
