/*
 *  7.16.   ,   .   
 *    :    , 
 * , ,  ,   .   
 *    , :
 *   -     ;
 *   -   ;
 *   -   .
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <regex.h>

#define LINE_BUFFER_LEN         256
#define NUM_FIELDS              5
#define NUM_FILTERED_FIELDS     3

const char *strUsageShort    = "Usage: %s [OPTION]... [FILE]\n\
Try `%s -h' for more information.\n";

const char *strUsageLong     = "Usage: %s [OPTION]... [FILE]\n\
Filter records from FILE (STDIN if no filename given) and print them on STDOUT.\n\
Example: %s -i -m '^ford' -c black car.db\n\
\n\
Filter selection and interpretation:\n\
  -i          toggle ignore case distinctions; can be supplied more than\n\
              once, toggles case distinctions for all subsequent filters;\n\
              initially case is honored\n\
  -p REGEX    add regular expression REGEX as car license plate filter\n\
  -c REGEX    add regular expression REGEX as car color filter\n\
  -m REGEX    add regular expression REGEX as car model filter\n\
\n\
Miscellaneous:\n\
  -h    display this help and exit\n\
\n\
Filter combinations:\n\
  More than one filter of the same type can be specified; in that case, all\n\
  the filters of the same type are OR'ed together.  For instance, options\n\
  `-c ^bl -c red' will select all cars of colors beginning with `bl' or\n\
  containing `red'.\n\
\n\
  Filters for different fields are AND'ed together.  For instance, options\n\
  `-i -c '^black$' -m ford' will select all black fords.\n\
\n\
Database file format:\n\
  Field 1: number on the license plate of the car\n\
  Field 2: color of the car\n\
  Field 3: brand and model of the car\n\
  Field 4: year of production of the car\n\
  Field 5: address of the owner of the car\n\
\n\
  Each field starts on a new line; one empty line separates records.\n";

const char *strOptions       = ":hip:c:m:";
const char *strErrOpenFile   = "%s: error opening file `%s'.\n";
const char *strMissingOptArg = "%s: option `-%c' requires an argument <%s>.\n";
const char *strUnknownOpt    = "%s: unknown option `-%c'.\n";
const char *strUnknownOptChar= "%s: unknown option character `\\x%x'.\n";
const char *strErrMem        = "%s: not enough memory.\n";
const char *strErrUnexpectedEOF = "%s: error reading input--unexpected end of file.\n";
      char *strProgName;

int RegexIgnoreCase = 0; /* initially honor the case */

typedef struct _filter {
  int       len;
  regex_t **filter;
} filter, *filterptr;

filterptr Filters;
char    **Fields;


int  SetUpBuffers(void);
int  ParseOptions(int argc, char **argv);
void PrintRegError(int RegErr, regex_t *compiled);


int main(int argc, char **argv)
{
  int   i, j;
  int   BailOut;
  char *ptrTemp;

  strProgName = *argv;

  if ( !SetUpBuffers() )
    return -1;

  if ( !ParseOptions(argc, argv) )
    return -1;

  /* Main loop */
  BailOut = 0;
  while ( !BailOut ) {
    int reject = 0;

    /* Read in all the fields of the current record */
    for ( i = 0; i < NUM_FIELDS; i++ ) {
      if ( !fgets(*(Fields + i), LINE_BUFFER_LEN, stdin) ) {
        if ( i != 0 )
          fprintf(stderr, strErrUnexpectedEOF, strProgName);
        BailOut = -1;
        break;
      }
      /* Remove trailing newline from the buffer */
      ptrTemp = *(Fields + i);
      while ( *ptrTemp )
        ptrTemp++;
      ptrTemp--;
      if ( *ptrTemp == '\n' )
        *ptrTemp = '\0';
    }
    /* Filter this record */
    if ( !BailOut ) {
      /* Check record against the filter */
      for ( i = 0; i < NUM_FILTERED_FIELDS; i++ ) {
        regex_t **ptrFilter = (Filters + i)->filter;
        int       satisfied = 0;
        for ( j = (Filters + i)->len - 1; j >= 0; j-- )
          if ( !regexec(*(ptrFilter + j), *(Fields + i), 0, NULL, 0) ) {
            satisfied = -1;
            break;
          }
        if ( !satisfied && (Filters + i)->len > 0 ) {
          reject = -1;
          break;
        }
      }
      /* Print the record if it passed through the filter */
      if ( !reject ) {
        for ( i = 0; i < NUM_FIELDS; i++ )
          printf("%s\n", *(Fields + i));
        putchar('\n');
      }
      /* Read in newline separating the records */
      fgets(*Fields, LINE_BUFFER_LEN, stdin);
    }
  }

  return 0;
}


int SetUpBuffers(void)
{
  int i;

  if ( !(Filters = malloc(NUM_FILTERED_FIELDS * sizeof( *Filters ))) ||
       !(Fields  = malloc(NUM_FIELDS          * sizeof( *Fields  ))) ) {
    fprintf(stderr, strErrMem, strProgName);
    return 0;
  }

  for ( i = 0; i < NUM_FIELDS; i++ )
    if ( !(*(Fields+i) = malloc(LINE_BUFFER_LEN)) ) {
      fprintf(stderr, strErrMem, strProgName);
      return 0;
    }

  for ( i = 0; i < NUM_FILTERED_FIELDS; i++ ) {
    (Filters + i)->len = 0;
    (Filters + i)->filter = NULL;
  }

  return -1;
}


int ParseOptions(int argc, char **argv)
{
  int       c, i, len, RegErr;
  regex_t **ptrFilter;
  char      *ptrTemp;

  opterr = 0; /* prevent GETOPT from printing error messages */

  while ( (c = getopt (argc, argv, strOptions)) != -1 ) {
    switch ( c ) {
    case 'h': /* help */
      fprintf(stdout, strUsageLong, strProgName, strProgName); 
      return 0;
    case 'i': /* toggle case insensitive regexp search */
      RegexIgnoreCase = (RegexIgnoreCase == 0) ? REG_ICASE : 0;
      continue;
      break;
    case 'p': /* license plate */
      i = 0;
      break;
    case 'c': /* color */
      i = 1;
      break;
    case 'm': /* model */
      i = 2;
      break;
    case '?': /* unknown option */
      if ( isprint (optopt) )
        fprintf(stderr, strUnknownOpt, strProgName, optopt);
      else
        fprintf(stderr, strUnknownOptChar, strProgName, optopt);
      fprintf(stderr, strUsageShort, strProgName, strProgName); 
      return 0;
      break;
    case ':': /* missing option argument */
      switch ( optopt ) {
      case 'p':
        ptrTemp = "license plate";
        break;
      case 'c':
        ptrTemp = "color";
        break;
      case 'm':
        ptrTemp = "model";
        break;
      default:
        abort(); /* should never get here */
      }
      fprintf(stderr, strMissingOptArg, strProgName, optopt, ptrTemp);
      return 0;
      break;
    default:
      abort(); /* should never get here */
    }
    /* Add new regexp to the filter */
    len = ++(Filters+i)->len;
    ptrFilter = (Filters+i)->filter;
    /* Grow filter array by one and allocate new regex structure */
    if ( !(ptrFilter = (regex_t **)
           realloc(ptrFilter, len * sizeof(*ptrFilter))) ||
         !(*(ptrFilter + len - 1) = (regex_t *) malloc(sizeof(regex_t))) ) {
      fprintf(stderr, strErrMem, strProgName);
      return 0;
    }
    (Filters+i)->filter = ptrFilter; /* save new location of filter array */
    /* Compile the regex right off */
    RegErr=regcomp(*(ptrFilter + len - 1), optarg, REG_NOSUB | RegexIgnoreCase);
    if ( RegErr ) {
      PrintRegError(RegErr, *(ptrFilter+len-1));
      return 0;
    }
  }

  /* If user provided filename, try to redirect stdin to that file */
  if ( optind < argc ) {
    if ( !freopen(*(argv+optind), "r", stdin) ) {
      fprintf(stderr, strErrOpenFile, strProgName, *(argv+optind));
      return 0;
    }
  }

  return -1;
}


void PrintRegError(int RegErr, regex_t *compiled)
{
  char *strErr;
  int len;

  len = regerror(RegErr, compiled, NULL, 0);
  if ( !(strErr = (char *) malloc(len)) ) {
    fprintf(stderr, strErrMem, strProgName);
    return;
  }
  regerror(RegErr, compiled, strErr, len);
  fputs(strProgName, stderr);
  fputc(':', stderr);
  fputc(' ', stderr);
  fputs(strErr, stderr);
  fputc('\n', stderr);
  free((void *) strErr);
  return;
}



/*
 * Local variables:
 * compile-command: "gcc 7_16.c && echo && ./a.out -c '^bl' -c red -i -m ford 7_16.in"
 * End:
 */
