/*
 * Non-recursive algorithm for calculating determinant of a matrix.
 *
 * NOTE:  breaks when user feeds 1x1 matrix.
 */
#include <stdio.h>
#include <stdlib.h>

#define SWAP(a, b, tmp)      tmp = a; a = b; b = tmp


const char *strErrOpenFile      = "error: cannont open file `%s'\n";
const char *strErrMem           = "error: not enough memory\n";
const char *strErrUnexpectedEOF = "error: unexpected end of input\n";

int      n;
double **matrix;
int     *permutation;
int     *withdrawn;

double   result;


int main(int argc, char **argv)
{
  int    i, j, tmp;
  int    interact = -1;
  int    transp_c;
  double prod;

  /* If user provided filename, try to redirect stdin to that file */
  if ( argc > 1 ) {
    if ( !freopen(*(argv+1), "r", stdin) ) {
      fprintf(stderr, strErrOpenFile, *(argv+1));
      return -1;
    }
    interact = 0;
  }

  /* Get matrix dimensions from user */
  if ( interact )
    fputs("Enter matrix dimension: ", stdout);
  if ( scanf("%d", &n) == EOF ) {
    fputs(strErrUnexpectedEOF, stderr);
    return -1;
  }

  /* Allocate buffers */
  if ( !( matrix = (double **) malloc( sizeof(double *) * n ) ) ) {
    fputs(strErrMem, stderr);
    return -1;
  }
  for ( i = 0; i < n; i++ ) {
    if ( !( *(matrix + i) = (double *) malloc( sizeof(double) * n) ) ) {
      fputs(strErrMem, stderr);
      return -1;
    }
  }
  if ( !( permutation = (int *) malloc( sizeof(*permutation) * n ) ) ) {
    fputs(strErrMem, stderr);
    return -1;
  }
  if ( !( withdrawn   = (int *) malloc( sizeof(*withdrawn) * n ) ) ) {
    fputs(strErrMem, stderr);
    return -1;
  }

  /* Get matrix elements from user */
  if ( interact )
    fputs("Enter matrix elements: ", stdout);

  for ( i = 0; i < n; i++ )
    for ( j = 0; j < n; j++ )
      if ( scanf("%lf", *(matrix + i) + j) == EOF ) {
        fputs(strErrUnexpectedEOF, stderr);
        return -1;
      }

  /* Print out the matrix */
  if ( interact ) {
    int n_minus_one = n - 1;
    for ( i = 0; i < n; i++ ) {
      for ( j = 0; j < n_minus_one; j++ )
        fprintf(stdout, "%6.2lf ", *(*(matrix + i) + j));
      fprintf(stdout, "%6.2lf", *(*(matrix + i) + j));
      fputc('\n', stdout);
    }
    putchar('\n');
  }

  /* Initialize the buffers */
  for ( i = 0; i < n; i++ ) {
    *(permutation + i)   = i;
    *(withdrawn     + i) = i;
  }

  /* Main loop.  We calculate the determinant by adding up products of all
     possible combinations of n elements, taken one from each column and one
     from each row, with each of the products getting corresponding sign.

     To achieve this, we first determine all permutations of first n natural
     numbers.  Now, with each permutation we get coordinates of elements for
     next product like this: a_{1 i_1}, a_{2 i_2}, ... , a_{n i_n}, where i_1,
     i_2, ... , i_n is the current permutation of first n natural numbers.

     The algorithm for searching the permutations is taken from
     `permutations.c' */
  j = n - 1;
  result = 0;
  transp_c = 0;

  while ( 1 ) {
    if ( j == n - 1 ) {
      if ( transp_c % 2 )
        prod = -1;
      else
        prod = 1;
      for ( i = 0; i < n; i++ )
        prod *= *(*(matrix + i) + *(permutation + i));
      result += prod;

      j--;
    }
    if ( *(withdrawn + j) == n - 1 ) {
      if ( j == 0 ) /* All done */
        break;

      /* Move the withdrawn element back into place */
      tmp                    = *(permutation + j);
      *(permutation + j)     = *(permutation + n - 1);
      *(permutation + n - 1) = tmp;

      *(withdrawn + j) = j;
      j--;
      transp_c--;
    } else {
      /* Move the withdrawn element back into place */
      tmp = *(permutation + *(withdrawn + j));
      if ( *(withdrawn + j) != j ) {
        *(permutation + *(withdrawn + j)) = *(permutation + j);
      } else
        transp_c++;

      (*(withdrawn + j))++;

      /* Withdraw next element */
      *(permutation + j)                = *(permutation + *(withdrawn + j));
      *(permutation + *(withdrawn + j)) = tmp;

      j = n - 1;
    }
  }

  /* Print the result */
  fprintf(stdout, "%lf\n", result);

  /* Free the mallocs */
  for ( i = 0; i < n; i++ )
    free( (void *) *(matrix + i) );
  free( (void *) matrix );
  free( (void *) permutation );
  free( (void *) withdrawn );

  return 0;
}



/*
 * Local variables:
 * compile-command: "gcc -Wall determinant.c && echo && ./a.out <<< \"3 1 -1 0 -1 2 3 -1 0 1\""
 * End:
 */
